#!/usr/bin/env python3.9
# coding:utf-8
# Copyright (C) 2024 All rights reserved.
# FILENAME:    ~~/src/acqua/core.py
# VERSION:     0.1.4
# CREATED:     2024-10-24 14:29
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Third-party packages ###
from click import group

### Local modules ###
from acqua.commands import auth, build, clean, dashboard, deploy, pull


@group
def cli() -> None:
  """acqua"""


cli.add_command(auth, "auth")
cli.add_command(build, "build")
cli.add_command(clean, "clean")
cli.add_command(dashboard, "dashboard")
cli.add_command(deploy, "deploy")
cli.add_command(pull, "pull")


if __name__ == "__main__":
  cli()



#!/usr/bin/env python3.9
# coding:utf-8
# Copyright (C) 2024 All rights reserved.
# FILENAME:    ~~/src/acqua/inlets/lagoon.py
# VERSION:     0.1.4
# CREATED:     2024-10-24 14:29
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from functools import reduce
from json import dumps
from re import Match, search
from typing import ClassVar, Dict, List, Optional

### Third-party packages ###
from blessed import Terminal
from blessed.keyboard import Keystroke
from docker.models.containers import Container
from pydantic import BaseModel, ConfigDict, StrictInt, StrictStr, TypeAdapter
from rich.layout import Layout
from rich.live import Live
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

### Local modules ###
from acqua.inlets.estuary import Estuary
from acqua.types import JsonrpcResponse, Validator


class Lagoon(BaseModel):
  model_config: ConfigDict = ConfigDict(arbitrary_types_allowed=True)  # type: ignore[misc]
  container_index: StrictInt = 0
  container_names: List[StrictStr] = []
  containers: List[Container] = []
  daemon: Container

  ### Split layouts ###
  body: ClassVar[Layout] = Layout(name="body", minimum_size=4, ratio=8, size=17)
  straits: ClassVar[Layout] = Layout(name="straits", size=20)
  footer: ClassVar[Layout] = Layout(name="footer", size=3)
  main: ClassVar[Layout] = Layout(size=72)
  pane: ClassVar[Layout] = Layout()
  sidebar: ClassVar[Layout] = Layout(size=24)
  estuary: ClassVar[Estuary] = Estuary(height=16, width=72)

  ### Terminal ###
  terminal: ClassVar[Terminal] = Terminal()

  def model_post_init(self, _) -> None:  # type: ignore[no-untyped-def]
    self.pane.split_row(self.sidebar, self.main)
    self.main.split_column(self.body, self.footer)
    self.sidebar.split_column(self.straits)

  def display(self) -> None:
    with self.terminal.cbreak(), self.terminal.hidden_cursor(), Live(
      self.pane, refresh_per_second=4, transient=True
    ):
      try:
        while True:
          container_name: str = self.container_names[self.container_index]
          keystroke: Keystroke = self.terminal.inkey(timeout=0.25)
          if keystroke.code == self.terminal.KEY_UP and self.container_index > 0:
            self.container_index -= 1
          elif (
            keystroke.code == self.terminal.KEY_DOWN
            and self.container_index < len(self.container_names) - 1
          ):
            self.container_index += 1
          elif keystroke in {"Q", "q"}:
            raise StopIteration

          container_rows: str = ""
          if self.container_index > 0:
            container_rows = "\n".join(self.container_names[: self.container_index])
            container_rows += f"\n[reverse]{self.container_names[self.container_index]}[reset]\n"
          else:
            container_rows = f"[reverse]{self.container_names[self.container_index]}[reset]\n"
          if self.container_index < len(self.container_names) - 1:
            container_rows += "\n".join(self.container_names[self.container_index + 1 :])
          self.pane["straits"].update(Panel(container_rows, title="domains"))
          container_name: str = self.container_names[self.container_index]

          ### Body ###
          body_table: Table = Table(expand=True, show_lines=True)
          body_table.add_column(container_name, "dark_sea_green bold")
          network: Optional[Match] = search(
            r"(?<=acqua)-(sui|sui-devnet|sui-testnet)", container_name
          )
          if network:
            data: dict = {
              "id": "chain-id",
              "jsonrpc": "2.0",
              "method": "sui_getTotalTransactionBlocks",
              "params": [],
            }
            chain_identifier: JsonrpcResponse[str] = TypeAdapter(
              JsonrpcResponse[str]
            ).validate_json(
              self.daemon.exec_run(
                f"""
                curl -sSL "http://localhost:9000" -H "Content-Type: application/json" -X POST --data-raw '{dumps(data)}'
                """
              ).output
            )
            data["method"] = "sui_getLatestCheckpointSequenceNumber"
            latest_checkpoint: JsonrpcResponse[int] = TypeAdapter(
              JsonrpcResponse[int]
            ).validate_json(
              self.daemon.exec_run(
                f"""
                curl -sSL "http://localhost:9000" -H "Content-Type: application/json" -X POST --data-raw '{dumps(data)}'
                """
              ).output
            )
            data["method"] = "suix_getReferenceGasPrice"
            reference_gas_price: JsonrpcResponse[str] = TypeAdapter(
              JsonrpcResponse[str]
            ).validate_json(
              self.daemon.exec_run(
                f"""
                curl -sSL "http://localhost:9000" -H "Content-Type: application/json" -X POST --data-raw '{dumps(data)}'
                """
              ).output
            )
            data["method"] = "suix_getValidatorsApy"
            validators_apy: JsonrpcResponse[Validator] = TypeAdapter(
              JsonrpcResponse[Validator]
            ).validate_json(
              self.daemon.exec_run(
                f"""
                curl -sSL "http://localhost:9000" -H "Content-Type: application/json" -X POST --data-raw '{dumps(data)}'
                """
              ).output
            )
            average_apy: float = reduce(
              lambda accumulated, validator_apy: accumulated + validator_apy.apy,
              validators_apy.result.apys,
              0.0,
            )
            body_table.add_row(
              Text.assemble(
                "\n",
                ("Chain", "bold"),
                "\n".ljust(15),
                ("Chain Identifier:".ljust(24), "green bold"),
                f"{chain_identifier.result}".rjust(16),
                "\n".ljust(15),
                ("Latest Checkpoint:".ljust(24), "bright_magenta bold"),
                f"{latest_checkpoint.result}".rjust(16),
                "\n".ljust(15),
                ("Reference Gas Price:".ljust(24), "cyan bold"),
                f"{reference_gas_price.result}".rjust(16),
                "\n".ljust(15),
                ("Average Validators APY:".ljust(24), "light_coral bold"),
                f"{average_apy}".rjust(16),
                "\n",
              )
            )
          else:
            body_table.add_row(self.estuary.renderable)
          self.pane["body"].update(body_table)

          ### Footer ###
          self.pane["footer"].update(
            Panel(
              Text.assemble(
                "Select:".rjust(16),
                (" ↑↓ ", "bright_magenta bold"),
                " " * 16,
                "Exit:".rjust(16),
                ("  Q ", "red bold"),
              )
            )
          )
      except StopIteration:
        print("If you cling to life, you live in fear of death.")


__all__ = ("Lagoon",)



#!/usr/bin/env python3.9
# coding:utf-8
# Copyright (C) 2024 All rights reserved.
# FILENAME:    ~~/src/acqua/commands/auth.py
# VERSION:     0.1.4
# CREATED:     2024-10-24 14:29
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from os import getenv, path
from re import findall
from typing import Dict, List

### Third-party packages ###
from click import argument, command, option
from docker import DockerClient, from_env
from docker.errors import DockerException
from rich import print as rich_print


@command
@argument("rpcuser", nargs=1)
@option("--bash", is_flag=True, help="Persist authentications in .bashrc", type=bool)
@option("--zsh", is_flag=True, help="Persist authentications in .zshrc", type=bool)
def auth(bash: bool, rpcuser: str, zsh: bool) -> None:
  """Persist authentications in desired run-control file."""
  client: DockerClient
  try:
    client = from_env()
    if not client.ping():
      raise DockerException
  except DockerException:
    rich_print("[red bold]Unable to connect to docker daemon.")
    return
  shell_select: Dict[str, bool] = {"/bin/bash": bash, "/bin/zsh": zsh}
  shell: str = getenv("SHELL", "/bin/bash")
  try:
    shell = next(filter(lambda value: value[1], shell_select.items()))[0]
  except StopIteration:
    pass
  # TODO: validate rpcuser
  rc_path: str = {"/bin/bash": "~/.bashrc", "/bin/zsh": "~/.zshrc"}[shell]
  authenticated: bool = False
  if getenv("ACQUA_AUTH_RPCUSER", None) is not None:
    authenticated = True
  else:
    with open(path.expanduser(rc_path), "r") as rc_readonly:
      found: List[str] = findall(r"export ACQUA_AUTH_RPCUSER=", rc_readonly.read())
      authenticated = len(found) != 0
  if authenticated:
    rich_print("[yellow]Already authenticated![reset]")
    return
  if rpcuser is not None:
    with open(path.expanduser(rc_path), "a") as rc_output:
      rc_output.write(f"export ACQUA_AUTH_RPCUSER={ rpcuser }\n")
  rich_print("[green]Success!")
  rich_print(f"In order to activate authentication, run the following command:")
  rich_print(f"    [blue]source { path.expanduser(rc_path) }[reset]")


__all__ = ("auth",)



#!/usr/bin/env python3.9
# coding:utf-8
# Copyright (C) 2024 All rights reserved.
# FILENAME:    ~~/src/acqua/types/validator.py
# VERSION:     0.1.4
# CREATED:     2024-11-07 14:00
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from typing import List

### Third-party packages ###
from pydantic import BaseModel, StrictFloat, StrictStr


class Apy(BaseModel):
  address: StrictStr
  apy: StrictFloat


class Validator(BaseModel):
  apys: List[Apy]
  epoch: str


__all__ = ("Apy", "Validator")



#!/usr/bin/env python3.9
# coding:utf-8
# Copyright (C) 2024 All rights reserved.
# FILENAME:    ~~/src/acqua/__init__.py
# VERSION:     0.1.4
# CREATED:     2024-10-24 14:29
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION: https://www.w3docs.com/snippets/python/what-is-init-py-for.html
#
# HISTORY:
# *************************************************************

### Local modules ###
from acqua.commands import build, clean, deploy, dashboard

__all__ = ("build", "clean", "dashboard", "deploy")

__version__ = "0.1.4"



#!/usr/bin/env python3.9
# coding:utf-8
# Copyright (C) 2024 All rights reserved.
# FILENAME:    ~~/src/acqua/commands/clean.py
# VERSION:     0.1.4
# CREATED:     2024-10-24 14:29
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from re import match
from typing import List

### Third-party packages ###
from click import command, option
from docker import DockerClient, from_env
from docker.errors import DockerException, NotFound
from docker.models.containers import Container
from docker.models.networks import Network
from rich import print as rich_print
from rich.progress import track

### Local modules ###
from acqua.configs import NETWORK


@command
@option("--inactive", help="Query inactive containers for removal.", is_flag=True, type=bool)
def clean(inactive: bool) -> None:
  """Remove all active "acqua-*" containers, drop network."""
  client: DockerClient
  try:
    client = from_env()
    if not client.ping():
      raise DockerException
  except DockerException:
    rich_print("[red bold]Unable to connect to docker daemon.")
    return

  outputs: List[str] = []
  containers: List[Container] = client.containers.list(all=inactive)
  for container in track(containers, f"Clean {('active','all')[inactive]} containers:".ljust(42)):
    container_name: None | str = container.name
    if container_name is None:
      continue
    if match(r"acqua-*", container_name) is not None:
      container.stop()
      container.remove(v=True)  # if `v` is true, remove associated volume
      outputs.append(f"<Container '{ container.name }'> removed.")
  try:
    network: Network = client.networks.get(NETWORK)
    network.remove()
    outputs.append(f"<Network '{ NETWORK }'> removed.")
  except NotFound:
    pass
  list(map(rich_print, outputs))


__all__ = ("clean",)



#!/usr/bin/env python3.9
# coding:utf-8
# Copyright (C) 2024 All rights reserved.
# FILENAME:    ~~/src/acqua/types/build_enum.py
# VERSION:     0.1.4
# CREATED:     2024-10-24 14:29
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from typing import Dict, Literal

from pydantic import BaseModel, StrictStr


class Build(BaseModel):
  instructions: Dict[int, StrictStr]
  platform: StrictStr = "linux/amd64"


BuildEnum = Literal["acqua-sui", "acqua-sui-devnet", "acqua-sui-testnet"]


__all__ = ("Build", "BuildEnum")



#!/usr/bin/env python3.9
# coding:utf-8
# Copyright (C) 2024 All rights reserved.
# FILENAME:    ~~/src/acqua/inlets/estuary.py
# VERSION:     0.1.4
# CREATED:     2024-10-24 14:29
# AUTHOR:      Sitt Guruvanich <aekazitt+github@gmail.com>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from math import pi, sin
from random import randint
from time import time
from typing import ClassVar, List

### Third-party packages ###
from apscheduler.schedulers.background import BackgroundScheduler
from pydantic import BaseModel, ConfigDict, StrictInt, StrictFloat
from rich.console import RenderableType
from rich.text import Text

BLOCK_ELEMENTS = [" ", "░", "▒", "▓", "█"]
THEME = [
  "#641401",
  "#E72802",
  "#FB6400",
  "#E48B09",
  "#FFEC2A",
  "#F6F9C0",
  "#FAFAFA",
  "#82C8FF",
  "#4078F0",
  "#0A60E1",
]


class Pixel(BaseModel):
  offset: StrictFloat
  temperature: StrictInt

  def draw(self, size_index: int) -> str:
    return Text(BLOCK_ELEMENTS[size_index], no_wrap=True, style=THEME[self.temperature]).markup


class Estuary(BaseModel):
  model_config: ConfigDict = ConfigDict(arbitrary_types_allowed=True)  # type: ignore[misc]
  chars: List[str] = []
  pixels: List[Pixel] = []

  ### Dimensions ###
  height: StrictInt
  width: StrictInt
  ### Split layouts ###
  scheduler: ClassVar[BackgroundScheduler] = BackgroundScheduler()

  def model_post_init(self, _) -> None:
    self.scheduler.add_job(self.update, "interval", seconds=0.25)
    self.scheduler.start()
    for y in range(self.height):
      temp = randint(0, len(THEME) - 1)
      for x in range(self.width):
        pixel = Pixel(
          temperature=temp, offset=(x + y) * (pi / 15)
        )  # Offset for wavy effect based on position
        self.pixels.append(pixel)

  @property
  def renderable(self) -> RenderableType:
    return Text.from_markup("".join(self.chars))

  def update(self) -> None:
    chars: List[str] = []
    for pixel in self.pixels:
      size_index = int((sin(time() + pixel.offset) + 1) / 2 * (len(BLOCK_ELEMENTS) - 1))
      chars.append(pixel.draw(size_index))
    self.chars = chars


__all__ = ("Estuary",)



#!/usr/bin/env python3.9
# coding:utf-8
# Copyright (C) 2024 All rights reserved.
# FILENAME:    ~~/src/acqua/commands/__init__.py
# VERSION:     0.1.4
# CREATED:     2024-10-24 14:29
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION: https://www.w3docs.com/snippets/python/what-is-init-py-for.html
#
# HISTORY:
# *************************************************************

### Local modules ###
from acqua.commands.auth import auth
from acqua.commands.build import build
from acqua.commands.clean import clean
from acqua.commands.dashboard import dashboard
from acqua.commands.deploy import deploy
from acqua.commands.pull import pull

__all__ = ("auth", "build", "clean", "dashboard", "deploy", "pull")



#!/usr/bin/env python3.9
# coding:utf-8
# Copyright (C) 2024 All rights reserved.
# FILENAME:    ~~/src/acqua/commands/pull.py
# VERSION:     0.1.4
# CREATED:     2024-10-24 14:29
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from typing import Dict, List, Set

### Third-party packages ###
from click import command, option
from docker import DockerClient, from_env
from docker.errors import DockerException
from docker.models.images import Image
from rich import print as rich_print
from rich.progress import track

### Local modules ###
from acqua.configs import BUILDS


@command
@option(
  "--devnet",
  is_flag=True,
  help="Pull acqua-sui-devnet image from GitHub container registry",
  type=bool,
)
@option(
  "--mainnet", is_flag=True, help="Pull acqua-sui image from GitHub container registry", type=bool
)
@option(
  "--testnet",
  is_flag=True,
  help="Pull acqua-sui-testnet image from GitHub container registry",
  type=bool,
)
def pull(
  devnet: bool,
  mainnet: bool,
  testnet: bool,
) -> None:
  """Pull daemon images from GitHub container registry."""
  client: DockerClient
  try:
    client = from_env()
    if not client.ping():
      raise DockerException
  except DockerException:
    rich_print("[red bold]Unable to connect to docker daemon.")
    return

  image_names: List[str] = list(
    map(
      lambda image: image.tags[0].split(":")[0],
      filter(lambda image: len(image.tags) != 0, client.images.list()),
    )
  )
  pull_select: Dict[str, bool] = {
    "acqua-sui": mainnet,
    "acqua-sui-devnet": devnet,
    "acqua-sui-testnet": testnet,
  }

  ### Checks if specified images had been built previously ###
  outputs: List[str] = []
  built: Set[str] = {tag for tag in BUILDS.keys() if pull_select[tag] and tag in image_names}
  outputs += map(lambda tag: f"<Image: '{tag}'> already exists in local docker images.", built)
  list(map(rich_print, outputs))

  outputs = []
  pullables: List[str] = [
    tag for tag in BUILDS.keys() if pull_select[tag] and tag not in image_names
  ]
  pull_count: int = len(pullables)
  if pull_count != 0:
    for pullable in track(pullables, "Pulling images from ghcr.io".ljust(35)):
      client.images.pull(f"ghcr.io/aekasitt/{ pullable }:latest")
      image: Image = client.images.get(f"ghcr.io/aekasitt/{ pullable }:latest")
      image.tag(pullable)
      outputs.append(f"<Image '{ pullable }'> pulled from registry.")
  list(map(rich_print, outputs))


__all__ = ("pull",)



#!/usr/bin/env python3.9
# coding:utf-8
# Copyright (C) 2024 All rights reserved.
# FILENAME:    ~~/src/acqua/types/difficulty_adjustment.py
# VERSION:     0.1.4
# CREATED:     2024-10-24 14:29
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Third-party packages ###
from pydantic import BaseModel, Field, StrictFloat, StrictInt


class DifficultyAdjustment(BaseModel):
  adjusted_time_average: StrictInt = Field(alias="adjustedTimeAvg")
  difficulty_change: StrictFloat = Field(alias="difficultyChange")
  estimated_retarget_date: StrictInt = Field(alias="estimatedRetargetDate")
  next_retarget_height: StrictInt = Field(alias="nextRetargetHeight")
  previous_retarget: StrictFloat = Field(alias="previousRetarget")
  remaining_blocks: StrictFloat = Field(alias="remainingBlocks")
  remaining_time: StrictFloat = Field(alias="remainingTime")
  time_average: StrictInt = Field(alias="timeAvg")
  time_offset: StrictInt = Field(alias="timeOffset")


__all__ = ("DifficultyAdjustment",)



#!/usr/bin/env python3.9
# coding:utf-8
# Copyright (C) 2024 All rights reserved.
# FILENAME:    ~~/src/acqua/types/blockchain_info.py
# VERSION:     0.1.4
# CREATED:     2024-10-24 14:29
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Third-party packages ###
from pydantic import BaseModel, StrictInt, StrictStr


class BlockchainInfo(BaseModel):
  blocks: StrictInt
  chain: StrictStr
  size_on_disk: StrictInt
  time: StrictInt


__all__ = ("BlockchainInfo",)



#!/usr/bin/env python3.9
# coding:utf-8
# Copyright (C) 2024 All rights reserved.
# FILENAME:    ~~/src/acqua/commands/deploy.py
# VERSION:     0.1.4
# CREATED:     2024-10-24 14:29
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from io import BytesIO
from re import match
from time import sleep
from tarfile import TarInfo, open as open_tarfile
from typing import Dict, List

### Third-party packages ###
from click import command, option
from docker import DockerClient, from_env
from docker.errors import APIError, DockerException
from docker.models.containers import Container
from rich import print as rich_print
from rich.progress import track
from yaml import dump

### Local modules ###
from acqua.configs import FULLNODES, NETWORK, SERVICES
from acqua.types import Chain, Service, ServiceName


@command
@option("--devnet", cls=Chain, is_flag=True, type=bool, variants=("mainnet", "testnet"))
@option("--mainnet", cls=Chain, is_flag=True, type=bool, variants=("devnet", "testnet"))
@option("--testnet", cls=Chain, is_flag=True, type=bool, variants=("devnet", "mainnet"))
def deploy(
  devnet: bool,
  mainnet: bool,
  testnet: bool,
) -> None:
  """Deploy cluster."""
  client: DockerClient
  try:
    client = from_env()
    if not client.ping():
      raise DockerException
  except DockerException:
    rich_print("[red bold]Unable to connect to docker daemon.")
    return

  ### Setup Node Network ###
  node_selector: Dict[ServiceName, bool] = {
    "acqua-sui": mainnet,
    "acqua-sui-devnet": devnet,
    "acqua-sui-testnet": testnet,
  }
  node_name: ServiceName = "acqua-sui"
  try:
    node_name = next(filter(lambda value: value[1], node_selector.items()))[0]
  except StopIteration:
    pass

  ### Setup Cluster Network ###
  try:
    client.networks.create(NETWORK, check_duplicate=True)
  except APIError:
    pass

  ### Launch middlewares ###
  middleware_selector = {"acqua-postgres": True}
  middlewares: Dict[ServiceName, Service] = {
    key: value
    for key, value in SERVICES.items()
    if value.service_type == "middleware"
    if middleware_selector[key]
  }
  for name, middleware in track(middlewares.items(), f"Deploy middleware services".ljust(42)):
    flags: List[str] = list(middleware.command.values())
    environment: List[str] = middleware.env_vars
    ports: Dict[str, int] = {p.split(":")[0]: int(p.split(":")[1]) for p in middleware.ports}
    client.containers.run(
      middleware.image,
      command=flags,
      detach=True,
      environment=environment,
      name=name,
      network=NETWORK,
      ports=ports,  # type: ignore
    )

  ### Launch Node Daemon ###
  node: Service = SERVICES[node_name]
  for _ in track(range(1), f"Deploy { node_name }".ljust(42)):
    flags: List[str] = list(node.command.values())
    ports: Dict[str, int] = {port.split(":")[0]: int(port.split(":")[1]) for port in node.ports}
    client.containers.run(
      node.image,
      command=flags,
      detach=True,
      environment=node.env_vars,
      name=node_name,
      network=NETWORK,
      ports=ports,  # type: ignore
    )

  ### Setup node peers ###
  daemon: Container
  try:
    daemon = next(
      filter(lambda acqua: match(r"acqua-(sui|sui-testnet)", acqua.name), client.containers.list())
    )
  except StopIteration:
    return
  sleep(3)  # delay until node running and fullnode.yaml generated
  extension_selector: str = ("mainnet", "testnet")[daemon.name == "acqua-sui-testnet"]
  extension: bytes = dump(FULLNODES[extension_selector].model_dump(by_alias=True)).encode("utf-8")
  fullnode_archive, stat = daemon.get_archive("/root/.sui/sui_config/fullnode.yaml")
  new_fullnode: BytesIO = BytesIO()
  with open_tarfile(fileobj=new_fullnode, mode="w|") as tar_file:
    content: bytes = b"".join([chunk for chunk in fullnode_archive])
    content += extension
    tar_info: TarInfo = TarInfo("fullnode.yaml")
    tar_info.size = stat.get("size", 0) + len(extension)
    tar_file.addfile(tar_info, BytesIO(content))
  new_fullnode.seek(0)
  daemon.put_archive(data=new_fullnode, path="/root/.sui/sui_config")


__all__ = ("deploy",)



#!/usr/bin/env python3.9
# coding:utf-8
# Copyright (C) 2024 All rights reserved.
# FILENAME:    ~~/src/acqua/types/jsonrpc_response.py
# VERSION:     0.1.4
# CREATED:     2024-11-07 02:00
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from typing import Generic, Literal, TypeVar

### Third-party packages ###
from pydantic import BaseModel, StrictInt, StrictStr


T = TypeVar("T")


class JsonrpcResponse(BaseModel, Generic[T]):
  jsonrpc: Literal["2.0"]
  id: StrictInt | StrictStr
  result: T


__all__ = ("JsonrpcResponse",)



#!/usr/bin/env python3.9
# coding:utf-8
# Copyright (C) 2024 All rights reserved.
# FILENAME:    ~~/src/acqua/types/blockchain_info.py
# VERSION:     0.1.4
# CREATED:     2024-10-24 14:29
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Third-party packages ###
from pydantic import BaseModel, Field, StrictBool, StrictInt, StrictFloat


class MempoolInfo(BaseModel):
  full_rbf: StrictBool = Field(
    alias="fullrbf",
    description="True if mempool accepts RBF without replaceability signaling introspection",
  )
  loaded: StrictBool = Field(description="True if mempool is fully loaded")
  max_mempool: StrictInt = Field(
    alias="maxmempool", description="Maxmimum memory usage for the mempool"
  )
  mempool_minimum_fee: StrictFloat = Field(
    alias="mempoolminfee", description="Minimum fee rate in BTC/kvB for transaction to be accepted"
  )
  minimum_relay_transaction_fee: StrictFloat = Field(
    alias="minrelaytxfee", description="Minimum relay fees for transaction"
  )
  total_fee: StrictFloat = Field(description="Total fees for the mempool in BTC")
  txn_count: StrictInt = Field(alias="size", description="Current transaction count")
  txn_bytes: StrictInt = Field(
    alias="bytes", description="Sum of all virtual transaction sizes as defined in BIP-141"
  )
  usage: StrictInt = Field(description="Total memory usage for mempool")
  unbroadcast_count: StrictInt = Field(
    alias="unbroadcastcount",
    description="Number of transactions that have not passed initial broadcast",
  )


__all__ = ("MempoolInfo",)



#!/usr/bin/env python3.9
# coding:utf-8
# Copyright (C) 2024 All rights reserved.
# FILENAME:    ~~/src/acqua/commands/build.py
# VERSION:     0.1.4
# CREATED:     2024-10-24 14:29
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from io import BytesIO
from typing import Dict, List, Set

### Third-party packages ###
from click import command, option
from docker import DockerClient, from_env
from docker.errors import BuildError, DockerException
from rich import print as rich_print

### Local modules ###
from acqua.configs import BUILDS
from acqua.inlets import Fjord
from acqua.types import Build


@command
@option("--devnet", is_flag=True, help="Build acqua-sui-devnet image", type=bool)
@option("--mainnet", is_flag=True, help="Build acqua-sui image", type=bool)
@option("--testnet", is_flag=True, help="Build acqua-sui-testnet image", type=bool)
def build(devnet: bool, mainnet: bool, testnet: bool) -> None:
  """Build peripheral images for the desired cluster."""
  client: DockerClient
  try:
    client = from_env()
    if not client.ping():
      raise DockerException
  except DockerException:
    rich_print("[red bold]Unable to connect to docker daemon.")
    return

  image_names: List[str] = list(
    map(
      lambda image: image.tags[0].split(":")[0],
      filter(lambda image: len(image.tags) != 0, client.images.list()),
    )
  )
  build_select: Dict[str, bool] = {
    "acqua-sui": mainnet,
    "acqua-sui-devnet": devnet,
    "acqua-sui-testnet": testnet,
  }

  ### Checks if specified images had been built previously ###
  outputs: List[str] = []
  built: Set[str] = {tag for tag in BUILDS.keys() if build_select[tag] and tag in image_names}
  outputs += map(lambda tag: f"<Image: '{tag}'> already exists in local docker images.", built)
  list(map(rich_print, outputs))

  builds: Dict[str, Build] = {
    tag: build for tag, build in BUILDS.items() if build_select[tag] and tag not in image_names
  }
  build_count: int = len(builds.keys())
  if build_count != 0:
    builds_items = builds.items()
    with Fjord(row_count=10) as fjord:
      task_id: int = fjord.add_task("", progress_type="primary", total=build_count)
      for tag, build in builds_items:
        build_task_id: int = fjord.add_task(tag, progress_type="build", total=100)
        with BytesIO("\n".join(build.instructions.values()).encode("utf-8")) as fileobj:
          try:
            fjord.progress_build(  # type: ignore[misc]
              client.api.build(
                container_limits={
                  "cpusetcpus": "0-3",
                  "cpushares": 4,
                  "memory": 8_589_934_592,  # 8GB
                  "memswap": 8_589_934_592,  # 8GB
                },
                decode=True,
                fileobj=fileobj,
                gzip=True,
                platform=build.platform,
                rm=True,
                shmsize=68_719_476_736,  # 64GB
                tag=tag,
              ),
              build_task_id,
            )
          except BuildError:
            fjord.update(build_task_id, completed=0)
          fjord.update(build_task_id, completed=100)
          fjord.update(task_id, advance=1)
      fjord.update(task_id, completed=build_count, description="[blue]Complete[reset]")


__all__ = ("build",)



#!/usr/bin/env python3.9
# coding:utf-8
# Copyright (C) 2024 All rights reserved.
# FILENAME:    ~~/src/acqua/inlets/fjord.py
# VERSION:     0.1.4
# CREATED:     2024-10-24 14:29
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from collections import deque
from math import floor
from re import search
from typing import Any, Deque, Generator, Optional, Union

### Third-party packages ###
from rich.box import MINIMAL
from rich.console import ConsoleRenderable, Group, RichCast
from rich.progress import BarColumn, Progress, Task
from rich.table import Table


class Fjord(Progress):
  primary_task: Task
  rows: Deque[str]
  table: Table = Table(box=MINIMAL, show_lines=False, show_header=False)

  def __init__(self, row_count: int) -> None:
    self.rows = deque(maxlen=row_count)
    super().__init__()

  def get_renderable(self) -> Union[ConsoleRenderable, RichCast, str]:
    return Group(self.table, *self.get_renderables())

  def get_renderables(self) -> Generator[Table, Any, Any]:
    for task in self.tasks:
      if task.fields.get("progress_type") == "build":
        image_name: str = task.description or "undefined"
        if task.completed <= 0:
          self.columns = (f"[red bold]Build unsuccessful for <Image '{image_name}'>.",)
        elif task.completed > 0 and task.completed < 100:
          self.columns = (
            f"Building <[bright_magenta]Image [green]'{image_name}'[reset]>…",
            "".ljust(9),
            BarColumn(),
          )
        else:
          self.columns = (
            f"[blue]Built <[bright_magenta]Image [green]'{image_name}'[reset]>[blue] successfully.[reset]",
          )
      elif task.fields.get("progress_type") == "primary":
        self.columns = ("Build specified images:".ljust(42), BarColumn())
      yield self.make_tasks_table([task])

  def progress_build(self, chunk: Generator, task_id: int) -> None:
    for line in chunk:
      if "stream" in line:
        stream: str = line.pop("stream").strip()
        step = search(r"^Step (?P<divided>\d+)\/(?P<divisor>\d+) :", stream)
        if step is not None:
          divided: int = int(step.group("divided"))
          divisor: int = int(step.group("divisor"))
          self.update(task_id, completed=floor(divided / divisor * 100))  # type: ignore
        self.update_table(stream)
      elif "error" in line:
        self.update_table(line.pop("error").strip())

  def update_table(self, row: Optional[str] = None) -> None:
    if row is not None:
      self.rows.append(row)
    table: Table = Table(box=MINIMAL, show_lines=False, show_header=False)
    for row_cell in self.rows:
      table.add_row(row_cell[0:92].ljust(92), style="grey50")
    self.table = table


__all__ = ("Fjord",)



#!/usr/bin/env python3.9
# coding:utf-8
# Copyright (C) 2024 All rights reserved.
# FILENAME:    ~~/src/acqua/configs.py
# VERSION:     0.1.4
# CREATED:     2024-10-24 14:29
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from pathlib import Path
from typing import Any, Dict, Literal, Optional

### Standard packages ###
from pydantic import TypeAdapter
from yaml import Loader, load

### Local modules ###
from acqua.types import Build, BuildEnum, Fullnode, Service, ServiceName

BUILDS: Dict[BuildEnum, Build]
FULLNODES: Dict[Literal["mainnet", "testnet"], Fullnode]
NETWORK: str
SERVICES: Dict[ServiceName, Service]


file_path: Path = Path(__file__).resolve()
with open(str(file_path).replace("configs.py", "fullnode.yml"), "rb") as stream:
  fullnodes: Optional[Dict[str, Any]] = load(stream, Loader=Loader)
  if fullnodes:
    FULLNODES = TypeAdapter(Dict[Literal["mainnet", "testnet"], Fullnode]).validate_python(
      fullnodes
    )
with open(str(file_path).replace("configs.py", "schemas.yml"), "rb") as stream:
  schema: Optional[Dict[str, Any]] = load(stream, Loader=Loader)
  if schema:
    BUILDS = TypeAdapter(Dict[BuildEnum, Build]).validate_python(schema["builds"])
    NETWORK = schema.get("network", "acqua")
    SERVICES = TypeAdapter(Dict[ServiceName, Service]).validate_python(schema["services"])

__all__ = ("BUILDS", "FULLNODE", "NETWORK", "SERVICES")



#!/usr/bin/env python3.9
# coding:utf-8
# Copyright (C) 2024 All rights reserved.
# FILENAME:    ~~/src/acqua/types/fullnode.py
# VERSION:     0.1.4
# CREATED:     2024-11-08 13:25
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from typing import List

### Third-party packages ###
from pydantic import BaseModel, Field, StrictStr


class SeedPeer(BaseModel):
  address: StrictStr
  peer_id: StrictStr = Field(alias="peer-id")


class P2pConfig(BaseModel):
  seed_peers: List[SeedPeer] = Field(alias="seed-peers")


class Fullnode(BaseModel):
  p2p_config: P2pConfig = Field(alias="p2p-config")


__all__ = ("Fullnode", "P2pConfig", "SeedPeer")



#!/usr/bin/env python3.9
# coding:utf-8
# Copyright (C) 2024 All rights reserved.
# FILENAME:    ~~/src/acqua/commands/dashboard.py
# VERSION:     0.1.4
# CREATED:     2024-10-24 14:29
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from re import match
from typing import List

### Third-party packages ###
from click import command
from docker import DockerClient, from_env
from docker.errors import DockerException
from docker.models.containers import Container
from rich import print as rich_print

### Local modules ###
from acqua.inlets import Lagoon


@command
def dashboard() -> None:
  """Dashboard for checking current state of images deployed."""
  client: DockerClient
  try:
    client = from_env()
    if not client.ping():
      raise DockerException
  except DockerException:
    rich_print("[red bold]Unable to connect to daemon.")
    return

  daemon: Container
  try:
    daemon = next(
      filter(
        lambda container: match(r"acqua-(sui|sui-devnet|sui-testnet)", container.name),
        reversed(client.containers.list()),
      )
    )
  except StopIteration:
    rich_print("[red bold] Cannot find active daemon.")
    return

  ### Retrieve other containers ###
  acqua_containers: List[Container] = list(
    filter(lambda container: match(r"acqua-*", container.name), reversed(client.containers.list()))
  )
  container_names: List[str] = [
    name for name in map(lambda container: container.name, acqua_containers) if name is not None
  ]
  lagoon: Lagoon = Lagoon(
    containers=acqua_containers,
    container_index=0,
    container_names=container_names,
    daemon=daemon,
  )
  lagoon.display()


__all__ = ("dashboard",)



#!/usr/bin/env python3.9
# coding:utf-8
# Copyright (C) 2024 All rights reserved.
# FILENAME:    ~~/src/acqua/types/__init__.py
# VERSION:     0.1.4
# CREATED:     2024-10-24 14:29
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION: https://www.w3docs.com/snippets/python/what-is-init-py-for.html
#
# HISTORY:
# *************************************************************

### Local modules ###
from acqua.types.blockchain_info import BlockchainInfo
from acqua.types.build import Build, BuildEnum
from acqua.types.chain import Chain
from acqua.types.difficulty_adjustment import DifficultyAdjustment
from acqua.types.fullnode import Fullnode
from acqua.types.jsonrpc_response import JsonrpcResponse
from acqua.types.mempool_info import MempoolInfo
from acqua.types.service import Service, ServiceName
from acqua.types.validator import Validator


__all__ = (
  "BlockchainInfo",
  "Build",
  "BuildEnum",
  "Chain",
  "DifficultyAdjustment",
  "Fullnode",
  "JsonrpcResponse",
  "MempoolInfo",
  "Service",
  "ServiceName",
  "Validator",
)



#!/usr/bin/env python3.9
# coding:utf-8
# Copyright (C) 2024 All rights reserved.
# FILENAME:    ~~/src/acqua/types/service.py
# VERSION:     0.1.4
# CREATED:     2024-10-24 14:29
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from typing import Dict, List, Literal
from typing_extensions import Annotated

### Third-party packages ###
from pydantic import BaseModel, Field, StrictStr

PortMapping = Annotated[StrictStr, Field(pattern=r"^\d{1,5}:\d{1,5}$")]


class Service(BaseModel):
  command: Dict[int, StrictStr] = {}
  env_vars: List[StrictStr] = []
  image: StrictStr
  ports: List[PortMapping]
  service_type: Literal["middleware", "node"] = Field(alias="type")


ServiceName = Literal["acqua-postgres", "acqua-sui", "acqua-sui-devnet", "acqua-sui-testnet"]

__all__ = ("PortMapping", "Service", "ServiceName")



#!/usr/bin/env python3.9
# coding:utf-8
# Copyright (C) 2024 All rights reserved.
# FILENAME:    ~~/src/acqua/types/chain.py
# VERSION:     0.1.4
# CREATED:     2024-10-24 14:29
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from typing import Any, List, Mapping, Tuple

### Third-party packages ###
from click import Context, Option, UsageError


class Chain(Option):
  def __init__(self, *args: Any, **kwargs: Any) -> None:
    self.variants: list = kwargs.pop("variants")
    assert self.variants, "'variants' parameter required."
    kwargs["help"] = (
      kwargs.get("help", "") + f"Option is mutually exclusive with {', '.join(self.variants)}."
    ).strip()
    super(Chain, self).__init__(*args, **kwargs)

  def handle_parse_result(
    self, ctx: Context, opts: Mapping[str, Any], args: List[str]
  ) -> Tuple[Any, List[str]]:
    current_opt: bool = self.name in opts
    for mutex_option in self.variants:
      if mutex_option in opts:
        if current_opt:
          raise UsageError(
            f"Illegal usage: '{self.name}' is mutually exclusive with '{mutex_option}'."
          )
        else:
          self.prompt = None
    return super(Chain, self).handle_parse_result(ctx, opts, args)


__all__ = ("Chain",)



#!/usr/bin/env python3.9
# coding:utf-8
# Copyright (C) 2024 All rights reserved.
# FILENAME:    ~~/src/acqua/inlets/__init__.py
# VERSION:     0.1.4
# CREATED:     2024-10-24 14:29
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION: https://www.w3docs.com/snippets/python/what-is-init-py-for.html
#
# HISTORY:
# *************************************************************

### Local modules ###
from acqua.inlets.estuary import Estuary
from acqua.inlets.fjord import Fjord
from acqua.inlets.lagoon import Lagoon

__all__ = ("Estuary", "Fjord", "Lagoon")
